/*
    FreeRTOS V6.0.5 - Copyright (C) 2010 Real Time Engineers Ltd.

    ***************************************************************************
    *                                                                         *
    * If you are:                                                             *
    *                                                                         *
    *    + New to FreeRTOS,                                                   *
    *    + Wanting to learn FreeRTOS or multitasking in general quickly       *
    *    + Looking for basic training,                                        *
    *    + Wanting to improve your FreeRTOS skills and productivity           *
    *                                                                         *
    * then take a look at the FreeRTOS eBook                                  *
    *                                                                         *
    *        "Using the FreeRTOS Real Time Kernel - a Practical Guide"        *
    *                  http://www.FreeRTOS.org/Documentation                  *
    *                                                                         *
    * A pdf reference manual is also available.  Both are usually delivered   *
    * to your inbox within 20 minutes to two hours when purchased between 8am *
    * and 8pm GMT (although please allow up to 24 hours in case of            *
    * exceptional circumstances).  Thank you for your support!                *
    *                                                                         *
    ***************************************************************************

    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    ***NOTE*** The exception to the GPL is included to allow you to distribute
    a combined work that includes FreeRTOS without being obliged to provide the
    source code for proprietary components outside of the FreeRTOS kernel.
    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/

/* Scheduler Includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"

/* Demo Includes. */
#include "serial.h"

/* Machine Includes. */
#ifdef __CPU__
#include __SFRFILE__(__CPU__)
#endif
/*---------------------------------------------------------------------------*/

/* Prototypes from the cstart.c file for enabling access to ENDINIT protected registers. */
extern void inline _endinit_clear(void);
extern void inline _endinit_set  (void);

/*
 * This reference is required by the Save/Restore Context Macros.
 */
extern volatile unsigned portBASE_TYPE * pxCurrentTCB;
/*-----------------------------------------------------------*/

/**
 * This function will check to see whether the Serial Transmit Interrupt is currently
 * activated, meaning that the interrupt is working through the back log of bytes
 * that it needs to send. If the ISR is not enabled, then it will be triggered to send
 * the first byte and it will be automatically re-triggered when that byte has been
 * sent. When the queue is exhausted, the ISR disables itself.
 * This function is privileged because it will trigger an interrupt.
 */
static void prvCheckTransmit( void ) PRIVILEGED_FUNCTION;
/*-----------------------------------------------------------*/

static xQueueHandle xSerialTransmitQueue = NULL;
static xQueueHandle xSerialReceiveQueue = NULL;
static volatile portBASE_TYPE xTransmitStatus = 0UL;
/*---------------------------------------------------------------------------*/

xComPortHandle xSerialPortInitMinimal( unsigned long ulWantedBaud, unsigned portBASE_TYPE uxQueueLength )
{
unsigned long ulReloadValue = 0UL;
	ulReloadValue = ( configCPU_CLOCK_HZ / ( 32 * ulWantedBaud ) ) - 1;

	if ( NULL == xSerialTransmitQueue )
	{
		xSerialTransmitQueue = xQueueCreate( uxQueueLength, sizeof( char ) );
		xSerialReceiveQueue = xQueueCreate( uxQueueLength, sizeof( char ) );
	}

	/* Enable ASC0 Module. */
	_endinit_clear();
		while ( 0 != ( WDT_CON0.U & 0x1UL ) );
		ASC0_CLC.U = 0x0200UL;
	_endinit_set();

	/* Disable the Operation. */
	ASC0_CON.U &= 0xFFFF7FFF;

	/* Set-up the GPIO Ports. */
	P5_IOCR0.U = 0x00009000;
	/* May need to configure P5_PDR.U for the pad mode. */

	/* Write the Baudrate. */
	ASC0_BG.U = ulReloadValue;

	/* Reconfigure and re-initialise the Operation. */
	ASC0_PISEL.U = 0UL;
	ASC0_CON.U = 0x00008011;	/* 1 Start, 1 Stop, 8 Data, No Parity, No Error Checking, Receive On, Module On. */

	/* Install the interrupts. */
	ASC0_TBSRC.U = 0x10UL | 0x5000UL;
	xTransmitStatus = 0UL;
	ASC0_RSRC.U = 0x14UL | 0x5000UL;

	/* COM Handle is never used by demo code. */
	return (xComPortHandle) pdPASS;
}
/*---------------------------------------------------------------------------*/

void vSerialPutString( xComPortHandle pxPort, const signed char * const pcString, unsigned short usStringLength )
{
unsigned short usChar;
	for ( usChar = 0; usChar < usStringLength; usChar++ )
	{
		(void)xSerialPutChar( pxPort, pcString[ usChar ], portMAX_DELAY );
	}
}
/*---------------------------------------------------------------------------*/

signed portBASE_TYPE xSerialGetChar( xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime )
{
	return xQueueReceive( xSerialReceiveQueue, pcRxedChar, xBlockTime );
}
/*---------------------------------------------------------------------------*/

signed portBASE_TYPE xSerialPutChar( xComPortHandle pxPort, signed char cOutChar, portTickType xBlockTime )
{
portBASE_TYPE xReturn = pdPASS;

	/* Send the character to the interrupt handler. */
	xReturn = xQueueSend( xSerialTransmitQueue, &cOutChar, xBlockTime );

	/* Start the transmission of bytes if necessary. */
	prvCheckTransmit();

	return xReturn;
}
/*---------------------------------------------------------------------------*/

void vSerialTransmitBufferInterrupt( void )
{
portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
unsigned char ucTx;

	/* ACK. */
	ASC0_TBSRC.U |= 0x4000UL;
	xTransmitStatus = 1UL;

	/* TBUF Can be refilled. */
	if ( pdPASS == xQueueReceiveFromISR( xSerialTransmitQueue, &ucTx, &xHigherPriorityTaskWoken ) )
	{
		ASC0_TBUF.U = ucTx;
	}
	else
	{
		/* Failed to get a character out of the Queue. No longer busy. */
		xTransmitStatus = 0UL;
	}
	portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
}
/*---------------------------------------------------------------------------*/

void vSerialReceiveInterrupt( void )
{
portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
unsigned char ucRx;

	/* Grab the character as early as possible. */
	ucRx = (unsigned char)ASC0_RBUF.U;

	/* ACK. */
	ASC0_RSRC.U |= 0x4000UL;

	/* Frame available in RBUF. */
	if ( pdPASS != xQueueSendFromISR( xSerialReceiveQueue, &ucRx, &xHigherPriorityTaskWoken ) )
	{
		/* Need some error handling code. */
	}
	portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
}
/*---------------------------------------------------------------------------*/

void prvCheckTransmit( void )
{
	/* Check to see if the interrupt handler is working its way through the buffer. */
	if ( 0 == xTransmitStatus )
	{
		/* Not currently operational so kick off the first byte. */
		ASC0_TBSRC.U |= 0x8000UL;
	}
}
/*---------------------------------------------------------------------------*/

#pragma section code "intvec.010"
void vSerialTransmitBufferInterruptHandler( void )
{
	portSAVE_CONTEXT();
	vSerialTransmitBufferInterrupt();
	portRESTORE_CONTEXT();
}
/*---------------------------------------------------------------------------*/

#pragma section code "intvec.014"
void vSerialReceiveInterruptHandler( void )
{
	portSAVE_CONTEXT();
	vSerialReceiveInterrupt();
	portRESTORE_CONTEXT();
}
/*---------------------------------------------------------------------------*/
