/*
    FreeRTOS V6.0.5 - Copyright (C) 2010 Real Time Engineers Ltd.

    ***************************************************************************
    *                                                                         *
    * If you are:                                                             *
    *                                                                         *
    *    + New to FreeRTOS,                                                   *
    *    + Wanting to learn FreeRTOS or multitasking in general quickly       *
    *    + Looking for basic training,                                        *
    *    + Wanting to improve your FreeRTOS skills and productivity           *
    *                                                                         *
    * then take a look at the FreeRTOS eBook                                  *
    *                                                                         *
    *        "Using the FreeRTOS Real Time Kernel - a Practical Guide"        *
    *                  http://www.FreeRTOS.org/Documentation                  *
    *                                                                         *
    * A pdf reference manual is also available.  Both are usually delivered   *
    * to your inbox within 20 minutes to two hours when purchased between 8am *
    * and 8pm GMT (although please allow up to 24 hours in case of            *
    * exceptional circumstances).  Thank you for your support!                *
    *                                                                         *
    ***************************************************************************

    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    ***NOTE*** The exception to the GPL is included to allow you to distribute
    a combined work that includes FreeRTOS without being obliged to provide the
    source code for proprietary components outside of the FreeRTOS kernel.
    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/

/* Standard includes. */
#include <stdlib.h>
#include <string.h>

/* Kernel includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "list.h"

/* Machine includes */
#ifdef __CPU__
#include __SFRFILE__(__CPU__)
#endif
/*-----------------------------------------------------------*/

/* System register Definitions. */
#define portSYSTEM_PROGRAM_STATUS_WORD					( (unsigned portBASE_TYPE) 0x000008FF ) /* Supervisor Mode, MPU Register Set 0 and Call Depth Counting disabled. */
#define portINITIAL_PRIVILEGED_PROGRAM_STATUS_WORD		( (unsigned portBASE_TYPE) 0x000014FF ) /* IO Level 1, MPU Register Set 1 and Call Depth Counting disabled. */
#define portINITIAL_UNPRIVILEGED_PROGRAM_STATUS_WORD	( (unsigned portBASE_TYPE) 0x000010FF ) /* IO Level 0, MPU Register Set 1 and Call Depth Counting disabled. */
#define portINITIAL_PCXI_UPPER_CONTEXT_WORD				( (unsigned portBASE_TYPE) 0x00C00000 )	/* The lower 20 bits identify the CSA address. */
#define portINITIAL_PCXI_LOWER_CONTEXT_WORD				( (unsigned portBASE_TYPE) 0x00000000 )	/* The lower 20 bits identify the CSA address. */
#define portUPPER_CONTEXT_BIT							( (unsigned portBASE_TYPE) 0x00400000 )	/* Bit that indicates whether the context is upper or lower. */

#if portUSING_MPU_WRAPPERS == 1
	#define portINITIAL_SYSCON							( (unsigned portBASE_TYPE) 0x00000002 )	/* MPU Enable. */
#else	/* portUSING_MPU_WRAPPERS != 1 */
	#define portINITIAL_SYSCON							( (unsigned portBASE_TYPE) 0x00000000 )	/* MPU Disable. */
#endif	/* portUSING_MPU_WRAPPERS */

#define portSELECT_PROGRAM_STATUS_WORD( xRunPrivileged )		( ( xRunPrivileged ) ? portINITIAL_PRIVILEGED_PROGRAM_STATUS_WORD : portINITIAL_UNPRIVILEGED_PROGRAM_STATUS_WORD )

/* CSA manipulation macros. */
#define portCSA_FCX_MASK					( 0x000FFFFFUL )

/* OS Interrupt and Trap mechanisms. */
#define portRESTORE_FIRST_CONTEXT()			( CPU_SRC0.U = ( portRESTORE_FIRST_TASK_PRIORITY_LEVEL ) | 0x00009001 )
#define portSYSCALL_TRAP					6

#define portSYSTEM_DATA_PRIVILEGES			( 0xC0C0C0C0UL )
#define portSYSTEM_CODE_PRIVILEGES			( 0x00008080UL )

#define portSYSTEM_PRIVILEGE_PROGRAM_STATUS_WORD	( (unsigned portBASE_TYPE) 0x00000800 ) /* Supervisor Mode. */
#define portMPU_SELECT_REGISTER_SET_MASK			( 0x00003000UL )						/* MPU Register Set Mask. */
#define portMPU_SELECT_REGISTER_SET_ONE				( 0x00001000UL )						/* MPU Register Set 1. */
/*-----------------------------------------------------------*/

#if configCHECK_FOR_STACK_OVERFLOW > 0
	#error "pxTopOfStack is used to store the last used CSA so it is not appropriate to enable stack checking."
	/* The stack pointer is accessible using portCSA_TO_ADDRESS( portCSA_TO_ADDRESS( pxCurrentTCB->pxTopOfStack )[ 0 ] )[ 2 ]; */
#endif /* configCHECK_FOR_STACK_OVERFLOW */
/*-----------------------------------------------------------*/

/*
 * This reference is required by the Save/Restore Context Macros.
 */
extern volatile unsigned portBASE_TYPE * pxCurrentTCB;
/*-----------------------------------------------------------*/

/*
 * Perform any hardware configuration necessary to generate the tick interrupt.
 */
static void prvSetupTimerInterrupt( void );

/**
 * Initialises the stack of the Task for MPU version.
 */
static portSTACK_TYPE *pxPortInitialiseStackMPU( portSTACK_TYPE * pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters, portBASE_TYPE xRunPrivileged );

#if portUSING_MPU_WRAPPERS == 1
	/**
	 * Sets the Kernel MPU Register Set.
	 */
	static void prvSetKernelMPURegisterSet( void ) PRIVILEGED_FUNCTION;

	/**
	 * Increases the IDLE task to have complete System access.
	 */
	static void prvElevateIdleTask( void ) PRIVILEGED_FUNCTION;
	/*-----------------------------------------------------------*/

	/**
	 * Wrapper to pxPortInitialiseStack() for the MPU port.
	 */
	portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE * pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters, portBASE_TYPE xRunPrivileged )
	{
		return pxPortInitialiseStackMPU( pxTopOfStack, pxCode, pvParameters, xRunPrivileged );
	}
	/*-----------------------------------------------------------*/
#else

	/* Define away the MPU specific functions. */
	#define prvSetKernelMPURegisterSet()
	#define prvElevateIdleTask()

	/**
	 * Wrapper to pxPortInitialiseStack() for the non-MPU port.
	 */
	portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE * pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters )
	{
		return pxPortInitialiseStackMPU( pxTopOfStack, pxCode, pvParameters, pdTRUE );
	}
	/*-----------------------------------------------------------*/
#endif

portSTACK_TYPE *pxPortInitialiseStackMPU( portSTACK_TYPE * pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters, portBASE_TYPE xRunPrivileged )
{
unsigned portBASE_TYPE *pxUpperCSA = NULL;
unsigned portBASE_TYPE *pxLowerCSA = NULL;

	/* 16 Address Registers (4 Address registers are global) and 16 Data Registers. */
	/* 3 System Registers */

	/* There are 3 registers that track the CSAs. */
	/* FCX points to the head of globally free set of CSAs.
	 * PCX for the task needs to point to Lower->Upper->NULL arrangement.
	 * LCX points to the last free CSA so that corrective action can be taken.
	 */

	/* Need two CSAs to store the context of a task.
	 * The upper context contains D8-D15, A10-A15, PSW and PCXI->NULL.
	 * The lower context contains D0-D7, A2-A7, A11 and PCXI->UpperContext.
	 * The pxCurrentTCB->pxTopOfStack points to the Lower Context ready for a restore Context.
	 * The Real stack pointer for the task is stored in the A10 which is restored with the upper context.
	 */

	/* Have to disable interrupts here because we are manipulating the CSAs. */
	portENTER_CRITICAL();
	{
		/* DSync to ensure that buffering is not a problem. */
		__dsync();

		/* Consume two Free CSAs. */
		pxLowerCSA = portCSA_TO_ADDRESS( __mfcr( FCX ) );
		if ( NULL != pxLowerCSA )
		{
			/* The Lower Links to the Upper. */
			pxUpperCSA = portCSA_TO_ADDRESS( pxLowerCSA[ 0 ] );
		}

		/* Check that we have successfully reserved two CSAs. */
		if ( ( NULL != pxLowerCSA ) && ( NULL != pxUpperCSA ) )
		{
			/* Remove the two consumed CSAs from the Free List. */
			__mtcr( FCX, pxUpperCSA[ 0 ] );
			/* ISync to commit the change to the FCX. */
			__isync();
		}
		else
		{
			/* For the time being, simply trigger a context list depletion trap. */
			__asm volatile( " SVLCX " );
		}
	}
	portEXIT_CRITICAL();

	/* Clear the CSA. */
	memset( pxUpperCSA, 0, 16 * sizeof( unsigned portBASE_TYPE ) );

	/* Upper Context. */
	pxUpperCSA[ 3 ] = (unsigned portBASE_TYPE)pxCode;					/* A11;	Return Address aka RA */
	pxUpperCSA[ 2 ] = (unsigned portBASE_TYPE)pxTopOfStack;				/* A10;	Stack Return aka Stack Pointer */
	pxUpperCSA[ 1 ] = portSELECT_PROGRAM_STATUS_WORD( xRunPrivileged );	/* PSW	*/
	/* PCXI - Including NULL pointer to indicate end of contexts. */
	pxUpperCSA[ 0 ] = portINITIAL_PCXI_UPPER_CONTEXT_WORD;

	/* Clear the CSA. */
	memset( pxLowerCSA, 0, 16 * sizeof( unsigned portBASE_TYPE ) );

	/* Lower Context. */
	pxLowerCSA[ 8 ] = (unsigned portBASE_TYPE)pvParameters;			/* A4;	Address Type Parameter Register	*/
	pxLowerCSA[ 1 ] = (unsigned portBASE_TYPE)pxCode;				/* A11;	Return Address aka RA */
	/* PCXI pointing to the Upper context. */
	pxLowerCSA[ 0 ] = ( portINITIAL_PCXI_UPPER_CONTEXT_WORD | (unsigned portBASE_TYPE)portADDRESS_TO_CSA( pxUpperCSA ) );

	/* Save the link to the CSA in the Top of Stack. */
	pxTopOfStack = (unsigned portBASE_TYPE *)portADDRESS_TO_CSA( pxLowerCSA );

	/* DSync to ensure that buffering is not a problem. */
	__dsync();

	return pxTopOfStack;
}
/*-----------------------------------------------------------*/

portBASE_TYPE xPortStartScheduler( void )
{
	/* Set-up the timer interrupt. */
	prvSetupTimerInterrupt();

	/* Configure the Kernel Registers for the MPU. */
	prvSetKernelMPURegisterSet();

	/* The IDLE task accesses privileged kernel data without proper privileges. */
	prvElevateIdleTask();

	/* Load the initial SYSCON. */
	__dsync();
	__mtcr( SYSCON, portINITIAL_SYSCON );

	/* ENDINIT has already been applied in the 'cstart.c' code. */

	/* Set-up the CPU_SRC0 to do the initial restore context. */
	portRESTORE_FIRST_CONTEXT();

	/* Enable interrupts to allow CPU_SRC0. */
	portENABLE_INTERRUPTS();

	portNOP();

	/* Should not get here. */
	return 0;
}
/*-----------------------------------------------------------*/

static void prvSetupTimerInterrupt( void )
{
    /* Set-up the Compare value. */
	STM_CMCON.U = ( 31UL - __clz( configCPU_CLOCK_HZ / configTICK_RATE_HZ ) );
	/* Take into account the current time so a tick doesn't happen immediately. */
	STM_CMP0.U = ( configCPU_CLOCK_HZ / configTICK_RATE_HZ ) + STM_TIM0.U;

	/* Set-up the interrupt. */
	STM_SRC0.U = ( portKERNEL_INTERRUPT_PRIORITY_LEVEL | 0x00005000UL );

	/* Enable the Interrupt. */
	STM_ISRR.U = 0x5UL;
	STM_ICR.U = 0x41UL;
}
/*-----------------------------------------------------------*/

/*
 * When a task is deleted, it is yielded permanently until the IDLE task
 * has an opportunity to reclaim the memory that that task was using.
 * Typically, the memory used by a task is the TCB and Stack but in the
 * TriCore this includes the CSAs that were consumed as part of the Call
 * Stack. These CSAs can only be returned to the Globally Free Pool when
 * they are not part of the current Call Stack, hence, delaying the
 * reclamation until the IDLE task is freeing the task's other resources.
 * This function uses the head of the linked list of CSAs (from when the
 * task yielded for the last time) and finds the tail (the very bottom of
 * the call stack) and inserts this list at the head of the Free list,
 * attaching the existing Free List to the tail of the reclaimed call stack.
 *
 * NOTE: the IDLE task needs processing time to complete this function
 * and in heavily loaded systems, the Free CSAs may be consumed faster
 * than they can be freed assuming that tasks are being spawned and
 * deleted frequently.
 */
void vPortReclaimCSA( unsigned portBASE_TYPE *pxTCB )
{
unsigned portBASE_TYPE pxHeadCSA, pxTailCSA, pxFreeCSA;

#if portUSING_MPU_WRAPPERS == 1
portBASE_TYPE xPriority = xPortRaisePrivilege();
#endif
	/* The first element in a TCB is the Last Used CSA.
	 * These simply need to be free'd to add them back to
	 * the global pool of CSAs.
	 */

	/* Lookup the first element from the TCB. */
	pxHeadCSA = ( *pxTCB ) & portCSA_FCX_MASK;

	/* If there is something to reclaim. */
	if ( 0UL != ( pxHeadCSA & portCSA_FCX_MASK ) )
	{
		/* Iterate over the CSAs that were consumed as part of the task. */
		pxTailCSA = pxHeadCSA;
		while ( 0UL != ( portCSA_TO_ADDRESS( pxTailCSA )[ 0 ] & portCSA_FCX_MASK ) )
		{
			/* Clear any extra bits from the link words. */
			portCSA_TO_ADDRESS( pxTailCSA )[ 0 ] = pxTailCSA & portCSA_FCX_MASK;

			/* Iterate to the next CSA. */
			pxTailCSA = portCSA_TO_ADDRESS( pxTailCSA )[ 0 ];
		}

		/* pxHeadCSA points to the first in the chain
		 * pxNextCSA points to the Head or the last in the chain.
		 */

		portENTER_CRITICAL();
		{
			/* Look up the current free CSA. */
			__dsync();
			pxFreeCSA = __mfcr( FCX );

			/* Join the current Free onto the Tail of what is being reclaimed. */
			portCSA_TO_ADDRESS( pxTailCSA )[ 0 ] = pxFreeCSA;

			/* Move the head of the reclaimed into the Free. */
			__dsync();
			__mtcr( FCX, pxHeadCSA );
			/* ISync to commit the change to the FCX. */
			__isync();
		}
		portEXIT_CRITICAL();
	}

#if portUSING_MPU_WRAPPERS == 1
	portMPU_RESTORE_PRIORITY( xPriority );
#endif
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* Nothing to do. Unlikely to want to end. */
}
/*-----------------------------------------------------------*/

#if portUSING_MPU_WRAPPERS == 1

void prvSetKernelMPURegisterSet( void )
{
	/* Set-up the Data regions for the Kernel. */
	__mtcr( DPR0_0L, (unsigned portBASE_TYPE)portRAM_START_ADDRESS );				/* All RAM. */
	__mtcr( DPR0_0U, (unsigned portBASE_TYPE)portRAM_STOP_ADDRESS );				/* All RAM. */
	__mtcr( DPR0_1L, (unsigned portBASE_TYPE)portINTERRUPT_STACK_START_ADDRESS );	/* Specifically Kernel Memory. */
	__mtcr( DPR0_1U, (unsigned portBASE_TYPE)portINTERRUPT_STACK_START_ADDRESS );	/* Specifically Kernel Memory. */
	__mtcr( DPR0_2L, (unsigned portBASE_TYPE)portCONTEXT_SAVE_AREA_START_ADDRESS );	/* Specifically CSAs. */
	__mtcr( DPR0_2U, (unsigned portBASE_TYPE)portCONTEXT_SAVE_AREA_STOP_ADDRESS );	/* Specifically CSAs. */
	__mtcr( DPR0_3L, (unsigned portBASE_TYPE)portROM_START_ADDRESS );				/* All ROM. */
	__mtcr( DPR0_3U, (unsigned portBASE_TYPE)portROM_STOP_ADDRESS );				/* All ROM. */

	/* Set-up the Code regions for the Kernel. */
	__mtcr( CPR0_0L, (unsigned portBASE_TYPE)portROM_START_ADDRESS );				/* entire flash region. */
	__mtcr( CPR0_0U, (unsigned portBASE_TYPE)portROM_STOP_ADDRESS );				/* entire flash region. */
	__mtcr( CPR0_1L, (unsigned portBASE_TYPE)portINTERRUPT_TABLE_START_ADDRESS );			/* Interrupts and Traps. */
	__mtcr( CPR0_1U, (unsigned portBASE_TYPE)portINTERRUPT_TABLE_STOP_ADDRESS );			/* Interrupts and Traps. */

	__mtcr( DPM0, (unsigned portBASE_TYPE)portSYSTEM_DATA_PRIVILEGES );
	__mtcr( CPM0, (unsigned portBASE_TYPE)portSYSTEM_CODE_PRIVILEGES );
}
/*-----------------------------------------------------------*/

/**
 * This function is called just before the scheduler starts. Therefore,
 * the IDLE task is guaranteed to be the least priority, last created
 * Task so it must be accessible from:
 * pxReadyTasksLists[ tskIDLE_PRIORITY ].pxIndex->pxPrevious->pvOwner;
 */
void prvElevateIdleTask( void )
{
extern xList pxReadyTasksLists[];
unsigned portBASE_TYPE *puxIdleTCB;
unsigned portBASE_TYPE *puxIdleContext;
	puxIdleTCB = pxReadyTasksLists[ tskIDLE_PRIORITY ].pxIndex->pvOwner;
	puxIdleContext = portCSA_TO_ADDRESS( puxIdleTCB[0] );		/* The lower context. */
	puxIdleContext = portCSA_TO_ADDRESS( puxIdleContext[0] );	/* The upper context. */
	puxIdleContext[ 1 ] = portSYSTEM_PROGRAM_STATUS_WORD;
}
/*-----------------------------------------------------------*/

/**
 * Switch MPU Register Set 1 to the values provided by pxMPUSettings.
 */
void vPortSetMPURegisterSetOne( xMPU_SETTINGS *pxMPUSettings )
{
	__asm volatile(												\
			" ld.d e14, [a4]0								\n"	\
			" mtcr #DPR1_0L, d14							\n" \
			" mtcr #DPR1_0U, d15							\n" \
			" ld.d e14, [a4]8								\n"	\
			" mtcr #DPR1_1L, d14							\n" \
			" mtcr #DPR1_1U, d15							\n" \
			" ld.d e14, [a4]16								\n"	\
			" mtcr #DPR1_2L, d14							\n" \
			" mtcr #DPR1_2U, d15							\n" \
			" ld.d e14, [a4]24								\n"	\
			" mtcr #DPR1_3L, d14							\n" \
			" mtcr #DPR1_3U, d15							\n" \
			" ld.d e14, [a4]32								\n"	\
			" mtcr #CPR1_0L, d14							\n" \
			" mtcr #CPR1_0U, d15							\n" \
			" ld.d e14, [a4]40								\n"	\
			" mtcr #CPR1_1L, d14							\n" \
			" mtcr #CPR1_1U, d15							\n" \
			" ld.d e14, [a4]48								\n"	\
			" mtcr #DPM1, d14								\n" \
			" mtcr #CPM1, d15								\n" \
			" isync											\n" \
			:													\
			:													\
			: "e14"											);
}
/*-----------------------------------------------------------*/

void vPortStoreTaskMPUSettings( xMPU_SETTINGS *pxMPUSettings, const struct xMEMORY_REGION * const pxRegions, portSTACK_TYPE *pxBottomOfStack, unsigned short usStackDepth )
{
	/* Reset the attributes for all Regions. */
	pxMPUSettings->uxDataRegionAttributes = 0UL;
	pxMPUSettings->uxCodeRegionAttributes = 0UL;

	if( pxRegions == NULL )
	{
		/* Default Settings for a task. */

		/* All RAM, all ROM, excluding kernel memory. */
		pxMPUSettings->xDataRegions[ 0 ].ulRegionLowerAddress = (unsigned long)portUNPRIVILEGED_RAM_START_ADDRESS;		/* Unprivileged RAM */
		pxMPUSettings->xDataRegions[ 0 ].ulRegionUpperAddress = (unsigned long)portUNPRIVILEGED_RAM_STOP_ADDRESS;		/* Unprivileged RAM */
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 0, portMPU_DATA_READ_WRITE_ACCESS );

		pxMPUSettings->xDataRegions[ 1 ].ulRegionLowerAddress = (unsigned long)portPRIVILEGED_RAM_START_ADDRESS;		/* Priviliged RAM */
		pxMPUSettings->xDataRegions[ 1 ].ulRegionUpperAddress = (unsigned long)portPRIVILEGED_RAM_STOP_ADDRESS;			/* Priviliged RAM */
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 1, portMPU_DATA_READ_ACCESS );

		pxMPUSettings->xDataRegions[ 2 ].ulRegionLowerAddress = (unsigned long)portUNPRIVILEGED_ROM_START_ADDRESS;		/* Unprivileged ROM */
		pxMPUSettings->xDataRegions[ 2 ].ulRegionUpperAddress = (unsigned long)portUNPRIVILEGED_ROM_STOP_ADDRESS;		/* Unprivileged RAM */
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 2, portMPU_DATA_READ_WRITE_ACCESS );

		pxMPUSettings->xDataRegions[ 3 ].ulRegionLowerAddress = (unsigned long)portPRIVILEGED_ROM_START_ADDRESS;		/* Privileged ROM */
		pxMPUSettings->xDataRegions[ 3 ].ulRegionUpperAddress = (unsigned long)portPRIVILEGED_ROM_STOP_ADDRESS;			/* Privileged ROM */
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 3, portMPU_DATA_READ_ACCESS );

		/* Code Regions. */
		pxMPUSettings->xCodeRegions[ 0 ].ulRegionLowerAddress = (unsigned long)portUNPRIVILEGED_ROM_START_ADDRESS;		/* Unprivileged ROM */
		pxMPUSettings->xCodeRegions[ 0 ].ulRegionUpperAddress = (unsigned long)portUNPRIVILEGED_ROM_STOP_ADDRESS;		/* Unprivileged ROM */
		pxMPUSettings->uxCodeRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 0, portMPU_CODE_EXECUTE_ACCESS );

		pxMPUSettings->xCodeRegions[ 1 ].ulRegionLowerAddress = 0UL;
		pxMPUSettings->xCodeRegions[ 1 ].ulRegionUpperAddress = 0UL;
		pxMPUSettings->uxCodeRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 1, portMPU_CODE_NO_EXECUTE );

	}
	else
	{

		/* This function is called automatically when the task is created - in
		which case the stack region parameters will be valid.  At all other
		times the stack parameters will not be valid and it is assumed that the
		stack region has already been configured. */
		if( usStackDepth > 0 )
		{
			/* Give access to the task Stack. */
			pxMPUSettings->xDataRegions[ 0 ].ulRegionLowerAddress = (unsigned long)pxBottomOfStack;
			pxMPUSettings->xDataRegions[ 0 ].ulRegionUpperAddress = (unsigned long)pxBottomOfStack + ( (unsigned long)usStackDepth * sizeof( portSTACK_TYPE ) );
		}
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 0, portMPU_DATA_READ_WRITE_ACCESS );

		/* Copy the regions specified by the xRegions Parameter into the task. */
		pxMPUSettings->xDataRegions[ 1 ].ulRegionLowerAddress = (unsigned long)(pxRegions[ 0 ].pvBaseAddress);
		pxMPUSettings->xDataRegions[ 1 ].ulRegionUpperAddress = (unsigned long)(pxRegions[ 0 ].pvBaseAddress) + pxRegions[ 0 ].ulLengthInBytes;
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 1, pxRegions[ 0 ].ulParameters );

		pxMPUSettings->xDataRegions[ 2 ].ulRegionLowerAddress = (unsigned long)(pxRegions[ 1 ].pvBaseAddress);
		pxMPUSettings->xDataRegions[ 2 ].ulRegionUpperAddress = (unsigned long)(pxRegions[ 1 ].pvBaseAddress) + pxRegions[ 1 ].ulLengthInBytes;
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 2, pxRegions[ 1 ].ulParameters );
		
		pxMPUSettings->xDataRegions[ 3 ].ulRegionLowerAddress = (unsigned long)(pxRegions[ 2 ].pvBaseAddress);
		pxMPUSettings->xDataRegions[ 3 ].ulRegionUpperAddress = (unsigned long)(pxRegions[ 2 ].pvBaseAddress) + pxRegions[ 2 ].ulLengthInBytes;
		pxMPUSettings->uxDataRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 3, pxRegions[ 2 ].ulParameters );

		/* Code Regions. */
		pxMPUSettings->xCodeRegions[ 0 ].ulRegionLowerAddress = (unsigned long)(pxRegions[ 3 ].pvBaseAddress);
		pxMPUSettings->xCodeRegions[ 0 ].ulRegionUpperAddress = (unsigned long)(pxRegions[ 3 ].pvBaseAddress) + pxRegions[ 3 ].ulLengthInBytes;
		pxMPUSettings->uxCodeRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 0, pxRegions[ 3 ].ulParameters );

		pxMPUSettings->xCodeRegions[ 1 ].ulRegionLowerAddress = (unsigned long)(pxRegions[ 4 ].pvBaseAddress);
		pxMPUSettings->xCodeRegions[ 1 ].ulRegionUpperAddress = (unsigned long)(pxRegions[ 4 ].pvBaseAddress) + pxRegions[ 4 ].ulLengthInBytes;
		pxMPUSettings->uxCodeRegionAttributes |= portMPU_SET_DATA_REGION_ATTRIBUTES( 1, pxRegions[ 4 ].ulParameters );

		/* The Regions have been configured within the TCB. Now we need to apply them. */
		if ( usStackDepth == 0 )
		{
			portENTER_CRITICAL();
				vPortSetMPURegisterSetOne( pxMPUSettings );
			portEXIT_CRITICAL();
		}
	}
}
/*-----------------------------------------------------------*/

portBASE_TYPE xPortRaisePrivilege( void )__attribute__((noinline))
{
unsigned portBASE_TYPE uxProgramStatusWord;

	__isync();
	uxProgramStatusWord = __mfcr( PSW );

	if ( 0 == ( uxProgramStatusWord & portSYSTEM_PRIVILEGE_PROGRAM_STATUS_WORD ) )
	{
		__asm volatile( " syscall #1 " );
	}
	return ( 0 == ( uxProgramStatusWord & portSYSTEM_PRIVILEGE_PROGRAM_STATUS_WORD ) );
}
/*-----------------------------------------------------------*/
#endif /* portUSING_MPU_WRAPPERS */

/*
 * Reloads the Timer ready for the next interrupt and increments the System Tick Count.
 * If using preemption, selects the next task to run.
 */
void vPortSystemTick( void )
{
	/* Clear the interrupt source. */
	STM_ISRR.U = 1UL;
	/* Reload the Compare Match register for X ticks into the future. */
	STM_CMP0.U += configCPU_CLOCK_HZ / configTICK_RATE_HZ;

	/* Increment the Tick. */
	vTaskIncrementTick();

#if configUSE_PREEMPTION == 1
	vTaskSwitchContext();
#endif
}
/*-----------------------------------------------------------*/

void vPortYield( unsigned portBASE_TYPE uxTrapIdentification )
{
unsigned portBASE_TYPE *puxCSAAddress;
	switch ( uxTrapIdentification )
	{
	case portSYSCALL_TASK_YIELD:
		/* Select another task to run. */
		vTaskSwitchContext();
		break;

	case portSYSCALL_RAISE_PRIORITY:
		puxCSAAddress = portCSA_TO_ADDRESS( *pxCurrentTCB );		/* Lower Context for xPortRaisePrivilege. */
		puxCSAAddress = portCSA_TO_ADDRESS( puxCSAAddress[ 0 ] );	/* Upper Context for xPortRaisePrivilege. */

		/* As long as xPortRaisePrivilege() is not inlined, then the Upper
		 * Context will always be the next one in the linked list. This is the
		 * context that we want to elevate. Note that as soon as that function
		 * is returned from, then the privileges will be restored.
		 */
		puxCSAAddress = portCSA_TO_ADDRESS( puxCSAAddress[ 0 ] );	/* Upper Context for MPU_Wrapper Function. */

		/* Select the Kernel mode MPU Register set and Give full access to all
		 * hardware and configuration registers (excluding ENDINIT protected).
		 */
		puxCSAAddress[ 1 ] &= ~portMPU_SELECT_REGISTER_SET_MASK;
		puxCSAAddress[ 1 ] |= portSYSTEM_PRIVILEGE_PROGRAM_STATUS_WORD;
		break;

	default:
		__debug();
		break;
	}
}
/*-----------------------------------------------------------*/

void __trap( portSYSCALL_TRAP ) vPortYieldHandler( void )
{
	portSAVE_CONTEXT();
	__asm volatile( " mov d4, d15 " );
	__asm volatile( " call vPortYield " );
	portRESTORE_CONTEXT();
}
/*-----------------------------------------------------------*/

#pragma section code "intvec.001"
void vRestoreFirstTaskContext( void )
{
	/* This handler is only used once to start the first task. */
	portRESTORE_CONTEXT();
}
/*-----------------------------------------------------------*/

#pragma section code "intvec.008"
void vPortSystemTickHandler( void )
{
	portSAVE_CONTEXT();
	vPortSystemTick();
	portRESTORE_CONTEXT();
}
/*-----------------------------------------------------------*/
